#include <unistd.h>

#include "BinaryBluetooth.h"
#include "bluetooth.h"
#include "hci.h"
#include "hci_lib.h"

bool v_BtConnect(std::string i_mac_address, int &o_handle)
{
    // open dev
    int dd = hci_power_on();
    if (dd < 0)
    {
        return false;
    }

    // connect device
    uint16_t handle = 0;
    uint8_t role = 0x00;
    unsigned int ptype = HCI_DM1 | HCI_DM3 | HCI_DM5 | HCI_DH1 | HCI_DH3 | HCI_DH5;
    bdaddr_t bdaddr;
    str2ba(i_mac_address.c_str(), &bdaddr);
    if (hci_create_connection(dd, &bdaddr, htobs(ptype), htobs(0x0000), role, &handle, 10000) < 0)
    {
        hci_close_dev(dd);
        return false;
    }
    o_handle = handle;

    // close dd
    hci_close_dev(dd);
    return true;
}

bool v_BtGetDeviceRssi(int i_handle, int &o_rssi)
{
    // open dev
    int dd = hci_power_on();
    if (dd < 0)
    {
        return false;
    }

    // get rssi
    int8_t rssi = 0;
    if (hci_read_rssi(dd, htobs(i_handle), &rssi, 1000) < 0)
    {
        hci_close_dev(dd);
        return false;
    }

    // close dd
    o_rssi = rssi;
    hci_close_dev(dd);
    return true;
}

bool v_BtDisconnect(int handle)
{
    // open dev
    int dd = hci_power_on();
    if (dd < 0)
    {
        return false;
    }

    // disconnect device
    bool hci_disconnect_success = true;
    if (hci_disconnect(dd, htobs(handle), HCI_OE_USER_ENDED_CONNECTION, 15000) != 0)
    {
        hci_disconnect_success = false;
    }

    // close and return value
    hci_close_dev(dd);
    if (!hci_disconnect_success)
    {
        return false;
    }
    return true;
}

BinaryBluetooth::BinaryBluetooth()
{
    AddNewCommand( "ReinitModule" ,   static_cast<CommandFunc>(&BinaryBluetooth::Pri_ReinitModule),    0 );
    AddNewCommand( "InterfaceUp",     static_cast<CommandFunc>(&BinaryBluetooth::InterfaceUp),         0 );
    AddNewCommand( "Scan",            static_cast<CommandFunc>(&BinaryBluetooth::Scan),                1 );
    AddNewCommand( "MacAddressGet",   static_cast<CommandFunc>(&BinaryBluetooth::MacAddressGet),       0 );
}

FuncRet BinaryBluetooth::InterfaceUp(std::vector<std::string> i_arg, std::vector<std::string> &o_msg)
{
    // open dev
    int dd = hci_power_on();
    if (dd < 0)
    {
        return BINARY_FUNC_RET_NG;
    }

    // interface up
    hci_close_dev(dd);
    return BINARY_FUNC_RET_OK;
}

FuncRet BinaryBluetooth::Scan(std::vector<std::string> i_arg, std::vector<std::string> &o_msg)
{
    int ant0_rssi = 0;
    if (i_arg[0].empty())
    {
        o_msg.push_back(std::to_string(ant0_rssi));
        return BINARY_FUNC_RET_NG;
    }
    std::string mac_address(i_arg[0]);

    printf("mac_address = %s\n", mac_address.c_str());
    // connect to device
    int handle = 0;
    if (!v_BtConnect(mac_address, handle))
    {
        o_msg.push_back(std::to_string(ant0_rssi));
        return BINARY_FUNC_RET_NG;
    }

    // getting rssi value
    bool flag_ng = true;
    if (v_BtGetDeviceRssi(handle, ant0_rssi))
    {
        o_msg.push_back(std::to_string(ant0_rssi));
        flag_ng = false;
    }
    
    // disconnect
    if (!v_BtDisconnect(handle))
    {
        printf("Unable to disconnect BT connection handle\n");
        o_msg.push_back(std::to_string(ant0_rssi));
        return BINARY_FUNC_RET_NG;
    }

    // return
    if (flag_ng)
    {
        o_msg.push_back(std::to_string(ant0_rssi));
        return BINARY_FUNC_RET_NG;
    }
    return BINARY_FUNC_RET_OK;
}

FuncRet BinaryBluetooth::MacAddressGet(std::vector<std::string> i_arg, std::vector<std::string> &o_msg)
{
    // get device
    int dd = hci_power_on();
    if (dd < 0)
    {
        printf("Unable to get BT device\n");
        return BINARY_FUNC_RET_NG;
    }

    // set request
    read_bd_addr_rp rp;
    struct hci_request rq;
    memset(&rq, 0, sizeof(rq));
    rq.ogf = OGF_INFO_PARAM;
    rq.ocf = OCF_READ_BD_ADDR;
    rq.rparam = &rp;
    rq.rlen = READ_BD_ADDR_RP_SIZE;

    // send request
    if (hci_send_req(dd, &rq, 100) < 0)
    {
        printf("Unable to send BT request\n");
        hci_close_dev(dd);
        return BINARY_FUNC_RET_NG;
    }

    // get bluetooth mac address
    bdaddr_t addr;
    char str_addr[18];
    bacpy(&addr, &rp.bdaddr);
    ba2str(&addr, str_addr);
    std::string mac_address(str_addr);
    o_msg.push_back(mac_address);

    // return
    hci_close_dev(dd);
    return BINARY_FUNC_RET_OK;
}

FuncRet BinaryBluetooth::Pri_ReinitModule(std::vector<std::string> i_arg, std::vector<std::string> &o_msg)
{
    return BINARY_FUNC_RET_OK;
}